package cz.muni.fi.pv112.cv2;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.joml.Matrix3f;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import org.lwjgl.BufferUtils;
import org.lwjgl.Version;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.MemoryStack;
import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;
import static org.lwjgl.system.MemoryStack.stackPush;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Cv2 {
    
    private static final int SIZEOF_AXES_VERTEX = 6 * Float.BYTES;
    private static final int COLOR_OFFSET = 3 * Float.BYTES;

    private static final float AXES[] = {
        // .. position .......... color ....
        // x axis
        1.0f, 0.0f, 0.0f,   1.0f, 0.0f, 0.0f,
        0.0f, 0.0f, 0.0f,   1.0f, 0.0f, 0.0f,
        // y axis
        0.0f, 1.0f, 0.0f,   0.0f, 1.0f, 0.0f,
        0.0f, 0.0f, 0.0f,   0.0f, 1.0f, 0.0f,
        // z axis
        0.0f, 0.0f, 1.0f,   0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, 0.0f,   0.0f, 0.0f, 1.0f,
    };

    private static final int SIZEOF_MODEL_VERTEX = 6 * Float.BYTES;
    private static final int NORMAL_OFFSET = 3 * Float.BYTES;

    private static final float CUBE[] = {
        // .. position ................ normal ....
        // front face
        -1.0f,  1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
        -1.0f, -1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
         1.0f, -1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
        -1.0f,  1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
         1.0f, -1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
         1.0f,  1.0f,  1.0f,     0.0f,  0.0f,  1.0f,
        // right face
         1.0f,  1.0f,  1.0f,     1.0f,  0.0f,  0.0f,
         1.0f, -1.0f,  1.0f,     1.0f,  0.0f,  0.0f,
         1.0f, -1.0f, -1.0f,     1.0f,  0.0f,  0.0f,
         1.0f,  1.0f,  1.0f,     1.0f,  0.0f,  0.0f,
         1.0f, -1.0f, -1.0f,     1.0f,  0.0f,  0.0f,
         1.0f,  1.0f, -1.0f,     1.0f,  0.0f,  0.0f,
        // back face
         1.0f,  1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
         1.0f, -1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
        -1.0f, -1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
         1.0f,  1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
        -1.0f, -1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
        -1.0f,  1.0f, -1.0f,     0.0f,  0.0f, -1.0f,
        // left face
        -1.0f,  1.0f, -1.0f,    -1.0f,  0.0f,  0.0f,
        -1.0f, -1.0f, -1.0f,    -1.0f,  0.0f,  0.0f,
        -1.0f, -1.0f,  1.0f,    -1.0f,  0.0f,  0.0f,
        -1.0f,  1.0f, -1.0f,    -1.0f,  0.0f,  0.0f,
        -1.0f, -1.0f,  1.0f,    -1.0f,  0.0f,  0.0f,
        -1.0f,  1.0f,  1.0f,    -1.0f,  0.0f,  0.0f,
        // top face
        -1.0f,  1.0f, -1.0f,     0.0f,  1.0f,  0.0f,
        -1.0f,  1.0f,  1.0f,     0.0f,  1.0f,  0.0f,
         1.0f,  1.0f,  1.0f,     0.0f,  1.0f,  0.0f,
        -1.0f,  1.0f, -1.0f,     0.0f,  1.0f,  0.0f,
         1.0f,  1.0f,  1.0f,     0.0f,  1.0f,  0.0f,
         1.0f,  1.0f, -1.0f,     0.0f,  1.0f,  0.0f,
        // bottom face
        -1.0f, -1.0f,  1.0f,     0.0f, -1.0f,  0.0f,
        -1.0f, -1.0f, -1.0f,     0.0f, -1.0f,  0.0f,
         1.0f, -1.0f, -1.0f,     0.0f, -1.0f,  0.0f,
        -1.0f, -1.0f,  1.0f,     0.0f, -1.0f,  0.0f,
         1.0f, -1.0f, -1.0f,     0.0f, -1.0f,  0.0f,
         1.0f, -1.0f,  1.0f,     0.0f, -1.0f,  0.0f,
    };
    
    // the window handle
    private long window;

    // window size
    private int width;
    private int height;
    private boolean resized = false;
    
    // animation
    private boolean animate = false;
    private float t = 0f;
    
    // rendering mode
    private int mode = GL_FILL;
    
    // model
    private ObjLoader teapot;
    
    // our OpenGL resources
    private int axesBuffer;
    private int cubeBuffer;
    private int teapotBuffer;
    private int axesArray;
    private int cubeArray;
    private int teapotArray;

    // our GLSL resources
    private int axesProgram;
    private int axesAspectUniformLoc;
    private int axesLengthUniformLoc;
    private int axesMvpUniformLoc;
    private int modelProgram;
    private int modelAspectUniformLoc;
    private int modelMvpUniformLoc;
    private int modelNUniformLoc;
    private int modelColorUniformLoc;
    
    public static void main(String[] args) {
        new Cv2().run();
    }
    
    public void run() {
        System.out.println("Hello LWJGL " + Version.getVersion() + "!");

        initGLFW();
        loop();

        // Free the window callbacks and destroy the window
        glfwFreeCallbacks(window);
        glfwDestroyWindow(window);

        // Terminate GLFW and free the error callback
        glfwTerminate();
        glfwSetErrorCallback(null).free();
    }

    private void initGLFW() {
	// Setup an error callback. The default implementation
        // will print the error message in System.err.
        GLFWErrorCallback.createPrint(System.err).set();

        // Initialize GLFW. Most GLFW functions will not work before doing this.
        if (!glfwInit()) {
            throw new IllegalStateException("Unable to initialize GLFW");
        }

        // Configure GLFW
        glfwDefaultWindowHints(); // optional, the current window hints are already the default
        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE); // the window will stay hidden after creation
        glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // the window will be resizable
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // request the OpenGL 3.3 core profile context
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

        // set initial width and height
        width = 640;
        height = 480;
        
        // Create the window
        window = glfwCreateWindow(width, height, "Hello World!", NULL, NULL);
        if (window == NULL) {
            throw new RuntimeException("Failed to create the GLFW window");
        }

        // Setup a key callback. It will be called every time a key is pressed, repeated or released.
        glfwSetKeyCallback(window, (window, key, scancode, action, mods) -> {
            if (action == GLFW_RELEASE) {
                switch (key) {
                    case GLFW_KEY_ESCAPE:
                        glfwSetWindowShouldClose(window, true); // We will detect this in the rendering loop
                        break;
                    // Task 7:  notice what this does
                    case GLFW_KEY_A:
                        animate = !animate;
                        break;
                    case GLFW_KEY_T:
                        // TODO toggle fullscreen
                        break;
                    // Task 5:  notice what is being done here (no need to change)
                    case GLFW_KEY_L:
                        mode = GL_LINE;
                        break;
                    case GLFW_KEY_F:
                        mode = GL_FILL;
                        break;
                }
            }
        });

        // add window size callback
        glfwSetWindowSizeCallback(window, (window, width, height) -> {
            this.width = width;
            this.height = height;
            resized = true;
        });
        
        // Get the thread stack and push a new frame
        try (MemoryStack stack = stackPush()) {
            IntBuffer pWidth = stack.mallocInt(1); // int*
            IntBuffer pHeight = stack.mallocInt(1); // int*

            // Get the window size passed to glfwCreateWindow
            glfwGetWindowSize(window, pWidth, pHeight);
            
            // Get the resolution of the primary monitor
            GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());

            // Center the window
            glfwSetWindowPos(
                    window,
                    (vidmode.width() - pWidth.get(0)) / 2,
                    (vidmode.height() - pHeight.get(0)) / 2
            );
        } // the stack frame is popped automatically

        // Make the OpenGL context current
        glfwMakeContextCurrent(window);
        // Enable v-sync
        glfwSwapInterval(1);

        // Make the window visible
        glfwShowWindow(window);
        
        // This line is critical for LWJGL's interoperation with GLFW's
        // OpenGL context, or any context that is managed externally.
        // LWJGL detects the context that is current in the current thread,
        // creates the GLCapabilities instance and makes the OpenGL
        // bindings available for use.
        GL.createCapabilities();
    }
    
    private void loop() {
        // Prepare data for rendering
        init();

        // Run the rendering loop until the user has attempted to close
        // the window or has pressed the ESCAPE key.
        while (!glfwWindowShouldClose(window)) {
            render();
            
            glfwSwapBuffers(window); // swap the color buffers

            // Poll for window events. The key callback above will only be
            // invoked during this call.
            glfwPollEvents();
        }
    }
    
    private void init() {
        // empty scene color
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glLineWidth(3.0f); // makes lines thicker
        
        // Task 5:  enable depth test in GL using glEnable(GL_DEPTH_TEST)
        glEnable(GL_DEPTH_TEST);
        
        // load GLSL program (vertex and fragment shaders)
        try {
            axesProgram = loadProgram("/resources/shaders/axes.vs.glsl",
                    "/resources/shaders/axes.fs.glsl");
            modelProgram = loadProgram("/resources/shaders/model.vs.glsl",
                    "/resources/shaders/model.fs.glsl");
        } catch (IOException ex) {
            Logger.getLogger(Cv2.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        
        // get uniform locations
        // axes program uniforms
        axesAspectUniformLoc = glGetUniformLocation(axesProgram, "aspect");
        axesLengthUniformLoc = glGetUniformLocation(axesProgram, "len");
        axesMvpUniformLoc = glGetUniformLocation(axesProgram, "MVP");
        // model program uniforms
        modelAspectUniformLoc = glGetUniformLocation(modelProgram, "aspect");
        modelMvpUniformLoc = glGetUniformLocation(modelProgram, "MVP");
        modelNUniformLoc = glGetUniformLocation(modelProgram, "N");
        modelColorUniformLoc = glGetUniformLocation(modelProgram, "color");

        // create buffers with geometry
        int[] buffers = new int[3];
        glGenBuffers(buffers);
        axesBuffer = buffers[0];
        cubeBuffer = buffers[1];
        teapotBuffer = buffers[2];

        // fill a buffers with geometry
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        glBufferData(GL_ARRAY_BUFFER, AXES, GL_STATIC_DRAW);

        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glBufferData(GL_ARRAY_BUFFER, CUBE, GL_STATIC_DRAW);

        // load teapot and fill buffer with teapot data
        teapot = new ObjLoader("/resources/models/teapot.obj");
        try {
            teapot.load();
        } catch (IOException ex) {
            Logger.getLogger(Cv2.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        int length = 3 * 6 * teapot.getTriangleCount();
        FloatBuffer teapot = BufferUtils.createFloatBuffer(length);
        for (int f = 0; f < this.teapot.getTriangleCount(); f++) {
            int[] pi = this.teapot.getVertexIndices().get(f);
            int[] ni = this.teapot.getNormalIndices().get(f);
            for (int i = 0; i < 3; i++) {
                float[] position = this.teapot.getVertices().get(pi[i]);
                float[] normal = this.teapot.getNormals().get(ni[i]);
                teapot.put(position);
                teapot.put(normal);
            }
        }
        teapot.rewind();
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glBufferData(GL_ARRAY_BUFFER, teapot, GL_STATIC_DRAW);
        
        // clear buffer binding, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        // create a vertex array object for the geometry
        int[] arrays = new int[3];
        glGenVertexArrays(arrays);
        axesArray = arrays[0];
        cubeArray = arrays[1];
        teapotArray = arrays[2];

        // get axes program attributes
        int positionAttribLoc = glGetAttribLocation(axesProgram, "position");
        int colorAttribLoc = glGetAttribLocation(axesProgram, "color");
        // bind axes buffer
        glBindVertexArray(axesArray);
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        // stride and offset are employed as both position and color data reside in the same vertex buffer
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, 0);
        glEnableVertexAttribArray(colorAttribLoc);
        glVertexAttribPointer(colorAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, COLOR_OFFSET);

        // get cube program attributes
        positionAttribLoc = glGetAttribLocation(modelProgram, "position");
        int normalAttribLoc = glGetAttribLocation(modelProgram, "normal");
        // bind cube buffer
        glBindVertexArray(cubeArray);
        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);
        
        // bind teapot buffer
        glBindVertexArray(teapotArray);
        glBindBuffer(GL_ARRAY_BUFFER, teapotBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);
        
        // clear bindings, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }
    
    private void render() {
        // Resize OpenGL viewport, i.e., the (bitmap) extents to that is the
        // OpenGL screen space [-1, 1] mapped.
        if (resized) {
            glViewport(0, 0, width, height);
            resized = false;
        }
        
        // animate variables
        if (animate) {
            t += 0.02f;
        }

        // Task 5:  change the constant GL_LINE to class variable mode
        glPolygonMode(GL_FRONT_AND_BACK, mode);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        
        Matrix4f transformation = new Matrix4f();
        // Task 1:  transform from Right Handed coordinate system to NDC,
        //              multiply Z by -1 using transformation.scale(float x, float y, float z)
        //          scale the cube to 50% of its size using transformation.scale(float xyz)
        // Task 2:  set scale of the scene to 25% so that we can see both cubes
        // Task 3:  project objects in the scene using perspective projection,
        //              use transformation.perspective((float) Math.toRadians(60f), (float) width / height, 1, 500)
        //          revert/remove the scale
        //          revert/remove the reflection of Z coordinate, remove multiplication of Z by -1
        //          move objects further by a small amount in -Z direction as the view frustum is [1, 500] in Z,
        //              use transformation.translate(float x, float y, float z)
//        transformation.scale(1f, 1f, -1f); // Task 1
//        transformation.scale(0.25f); // Task 1: 0.5f, Task 2: 0.25f
//        transformation.translate(0f, 0f, -5f); // Task 3
        

        // Task 4:  decouple transformation matrix into projection and view matrices
        //          create matrix Matrix4f projection and set it using
        //              projection.perspective((float) Math.toRadians(60f), width / (float) height, 1, 500)
        //          create matrix Matrix4f view and set it using view.lookAt(3f, 4f, 5f,
        //                                                                       0f, 0f, 0f,
        //                                                                       0f, 1f, 0f)
        Matrix4f projection = new Matrix4f()
                .perspective((float) Math.toRadians(60f), width / (float) height, 1, 500); // projection

        Matrix4f view = new Matrix4f()
                .lookAt(3f, 4f, 5f,
                        0f, 0f, 0f,
                        0f, 1f, 0f);

        // Task 4:  use matrices projection and view to draw the first cube,
        //              use drawModel(Matrix4f modelView, Matrix4f projection, int vao, int count, Vector3f color)
        //              use view matrix for "modelView" parameter, we don't want any model transformations
//        drawModel(view, projection, cubeArray, 36, new Vector3f(1f, 1f, 1f)); // Task 1: use transformation, Task 2: use view, projection

        // Task 2:  add another cube to scene so that its center lies at [-2.5, 0.0, -2.5],
        //              use transformation.translate(float x, float y, float z)
        // Task 4:  create Matrix4f modelView matrix, set it with view matrix
        //              and then add a translation using modelView.translate(...) to [-2.5, 0.0, -2.5]
        //          use matrices projection and modelView to draw the second cube,
        //              use drawModel(Matrix4f modelView, Matrix4f projection, int vao, int count, Vector3f color)
//        transformation.translate(-2.5f, 0f, -2.5f); // Task 2
//        Matrix4f modelView = new Matrix4f(view) // Task 4
//                .translate(-2.5f, 0f, -2.5f);
//        drawModel(modelView, projection, cubeArray, 36, new Vector3f(1f, 1f, 1f)); // Task 1: use transformation, Task 2: use modelView, projection
        
        // Task 6:  draw a table, use (5x) method drawTablePart(...) with following params,
        //              scale: (1.5f, 0.25f, 1.5f), position: (0f, 1.25f, 0f)
        //              scale: (0.25f, 1f, 0.25f), position: ( 1f, 0f,  1f)
        //              scale: (0.25f, 1f, 0.25f), position: (-1f, 0f,  1f)
        //              scale: (0.25f, 1f, 0.25f), position: ( 1f, 0f, -1f)
        //              scale: (0.25f, 1f, 0.25f), position: (-1f, 0f, -1f)
        drawTablePart(view, projection, new Vector3f(1.5f, 0.25f, 1.5f), new Vector3f(0f, 1.25f, 0f));
        drawTablePart(view, projection, new Vector3f(0.25f, 1f, 0.25f), new Vector3f(1f, 0f, 1f));
        drawTablePart(view, projection, new Vector3f(0.25f, 1f, 0.25f), new Vector3f(-1f, 0f, 1f));
        drawTablePart(view, projection, new Vector3f(0.25f, 1f, 0.25f), new Vector3f(1f, 0f, -1f));
        drawTablePart(view, projection, new Vector3f(0.25f, 1f, 0.25f), new Vector3f(-1f, 0f, -1f));

        // Task 6:  draw a teapot on top of the table
        //          create Matrix4f teapotMV matrix in order to place the teapot on the table,
        //              copy view matrix and update it with teapotMV.translate(...) to position [0f, 1.5f, 0f]
        //              scale the teapot to 50% of its size using teapotMV.scale(...)
        //          use draw model with appropriate vao
        // Task 7:  animate the teapot, multiply by rotation matrix of angle t around Y axis using
        //              teapotMV.rotate(float ang, float x, float y, float z)
        Matrix4f teapotMV = new Matrix4f(view)
                .translate(0f, 1.5f, 0f)
                .scale(0.5f)
                .rotate(t, 0f, 1f, 0f);

        drawModel(teapotMV, projection, teapotArray, teapot.getTriangleCount() * 3, new Vector3f(0.9f, 0.9f, 0.9f));
        
        // Task 8:  draw a yellow book (box) such that it lies on the closest corner of the table
        //          and it is rotated by 60 degrees around Y axis, use the following params
        //              position = [1f, 1.55f, 1f], rotation = 60° around Y, scale = (0.25f, 0.05f, 0.15f)
        //              NOTE, think about the order of transformations
        //              TIP, you can use drawAxes method to draw the local coordinate frame
        Matrix4f modelView = new Matrix4f(view)
                .translate(1f, 1.55f, 1f)
                .rotate((float) Math.toRadians(60.0), 0f, 1f, 0f)
                .scale(0.25f, 0.05f, 0.15f);

        drawModel(modelView, projection, cubeArray, 36, new Vector3f(1.0f, 1.0f, 0.0f));
        drawAxes(new Matrix4f(projection).mul(modelView), 3f);
    }

    private void drawModel(Matrix4f transformation, int vao, int count, Vector3f color) {
        drawModel(new Matrix4f().identity(), transformation, vao, count, color);
    }
    
    private void drawModel(Matrix4f modelView, Matrix4f projection, int vao, int count, Vector3f color) {
        // compute model-view-projection matrix
        Matrix4f mvp = new Matrix4f()
                .set(projection)
                .mul(modelView);
        
        // compute normal matrix
        Matrix3f n = modelView.get3x3(new Matrix3f())
                .invert()
                .transpose();
        
        glUseProgram(modelProgram);
        glBindVertexArray(vao); // bind vertex array to draw
        
        glUniform1f(modelAspectUniformLoc, width / (float) height); // update aspect uniform
        
        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        FloatBuffer nData = BufferUtils.createFloatBuffer(9);
        mvp.get(mvpData);
        n.get(nData);
        glUniformMatrix4fv(modelMvpUniformLoc, false, mvpData); // pass MVP matrix to shader
        glUniformMatrix3fv(modelNUniformLoc, false, nData); // pass Normal matrix to shader
        glUniform3f(modelColorUniformLoc, color.x, color.y, color.z); // pass object color to shader

        glDrawArrays(GL_TRIANGLES, 0, count);

        glBindVertexArray(0);
        glUseProgram(0);
    }
    
    private void drawTablePart(Matrix4f view, Matrix4f projection, Vector3f scale, Vector3f position) {
        Matrix4f modelView = new Matrix4f(view)
                .translate(position)
                .scale(scale);

        drawModel(modelView, projection, cubeArray, 36, new Vector3f(0.8f, 0.5f, 0.3f));
    }
    
    private void drawAxes(Matrix4f modelViewProjection, float length) {
        glUseProgram(axesProgram);
        glBindVertexArray(axesArray);
        
        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        modelViewProjection.get(mvpData);
        glUniform1f(axesLengthUniformLoc, length);
        glUniform1f(axesAspectUniformLoc, width / (float) height);
        glUniformMatrix4fv(axesMvpUniformLoc, false, mvpData);

        glDrawArrays(GL_LINES, 0, 6);

        glBindVertexArray(0);
        glUseProgram(0);
    }
    
    private int loadShader(String filename, int shaderType) throws IOException {
        String source = readAllFromResource(filename);
        int shader = glCreateShader(shaderType);
        
        // create and compile GLSL shader
        glShaderSource(shader, source);
        glCompileShader(shader);
        
        // check GLSL shader compile status
        int status = glGetShaderi(shader, GL_COMPILE_STATUS);
        if (status == GL_FALSE) {
            String log = glGetShaderInfoLog(shader);
            System.err.println(log);
        }
        
        return shader;
    }
    
    private int loadProgram(String vertexShaderFile, String fragmentShaderFile) throws IOException {
        // load vertex and fragment shaders (GLSL)
	int vs = loadShader(vertexShaderFile, GL_VERTEX_SHADER);
	int fs = loadShader(fragmentShaderFile, GL_FRAGMENT_SHADER);
        
	// create GLSL program, attach shaders and compile it
	int program = glCreateProgram();
	glAttachShader(program, vs);
	glAttachShader(program, fs);
	glLinkProgram(program);
        
        int status = glGetProgrami(program, GL_LINK_STATUS);
        if (status == GL_FALSE) {
            String log = glGetProgramInfoLog(program);
            System.err.println(log);
        }
        
        return program;
    }
    
    private String readAllFromResource(String resource) throws IOException {
        InputStream is = Cv2.class.getResourceAsStream(resource);
        if (is == null) {
            throw new IOException("Resource not found: " + resource);
        }
        
        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        StringBuilder sb = new StringBuilder();
        
        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }
        
        return sb.toString();
    }
    
}
